/*
 * Copyright (c) 2016, Intel Corporation
 * Authors: Salvatore Benedetto <salvatore.benedetto@intel.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public Licence
 * as published by the Free Software Foundation; either version
 * 2 of the Licence, or (at your option) any later version.
 */
#include <linux/kernel.h>
#include <linux/export.h>
#include <linux/err.h>
#include <linux/string.h>
#include <crypto/dh.h>
#include <crypto/kpp.h>

#undef dh
#undef crypto_dh_key_len
#define DH_KPP_SECRET_MIN_SIZE_NO_Q (sizeof(struct kpp_secret) + 3 * sizeof(int))

static inline int dh_data_size(const struct dh *p)
{
	return p->key_size + p->p_size + p->g_size;
}

int crypto_dh_key_len(const struct dh *p)
{
	return DH_KPP_SECRET_MIN_SIZE_NO_Q + dh_data_size(p);
}
EXPORT_SYMBOL_GPL(crypto_dh_key_len);

/* Old calculation with new structure */
static inline int dh_data_size_no_q(const struct dh_q *p)
{
	return p->key_size + p->p_size + p->g_size;
}

int crypto_dh_key_len_no_q(const struct dh_q *p)
{
	return DH_KPP_SECRET_MIN_SIZE_NO_Q + dh_data_size_no_q(p);
}

#define DH_KPP_SECRET_MIN_SIZE (sizeof(struct kpp_secret) + 4 * sizeof(int))

static inline u8 *dh_pack_data(void *dst, const void *src, size_t size)
{
	memcpy(dst, src, size);
	return dst + size;
}

static inline const u8 *dh_unpack_data(void *dst, const void *src, size_t size)
{
	memcpy(dst, src, size);
	return src + size;
}

static inline int dh_data_size_q(const struct dh_q *p)
{
	return p->key_size + p->p_size + p->q_size + p->g_size;
}

int crypto_dh_key_len_q(const struct dh_q *p)
{
	return DH_KPP_SECRET_MIN_SIZE + dh_data_size_q(p);
}
EXPORT_SYMBOL_GPL(crypto_dh_key_len_q);

int crypto_dh_encode_key(char *buf, unsigned int len, const struct dh_q *params)
{
	u8 *ptr = buf;
	bool do_q = true;
	struct kpp_secret secret = {
		.type = CRYPTO_KPP_SECRET_TYPE_DH,
		.len = len
	};

	if (unlikely(!buf))
		return -EINVAL;

	if (len == crypto_dh_key_len_no_q(params) && !params->q_size)
		do_q = false;
	else if (len != crypto_dh_key_len_q(params))
		return -EINVAL;

	ptr = dh_pack_data(ptr, &secret, sizeof(secret));
	ptr = dh_pack_data(ptr, &params->key_size, sizeof(params->key_size));
	ptr = dh_pack_data(ptr, &params->p_size, sizeof(params->p_size));
	if (do_q)
		ptr = dh_pack_data(ptr, &params->q_size, sizeof(params->q_size));
	ptr = dh_pack_data(ptr, &params->g_size, sizeof(params->g_size));
	ptr = dh_pack_data(ptr, params->key, params->key_size);
	ptr = dh_pack_data(ptr, params->p, params->p_size);
	if (do_q)
		ptr = dh_pack_data(ptr, params->q, params->q_size);
	dh_pack_data(ptr, params->g, params->g_size);

	return 0;
}
EXPORT_SYMBOL_GPL(crypto_dh_encode_key);

int crypto_dh_decode_key(const char *buf, unsigned int len, struct dh_q *params)
{
	const u8 *ptr = buf;
	struct kpp_secret secret;

	if (unlikely(!buf || len < DH_KPP_SECRET_MIN_SIZE_NO_Q))
		return -EINVAL;

	ptr = dh_unpack_data(&secret, ptr, sizeof(secret));
	if (secret.type != CRYPTO_KPP_SECRET_TYPE_DH)
		return -EINVAL;

	ptr = dh_unpack_data(&params->key_size, ptr, sizeof(params->key_size));
	ptr = dh_unpack_data(&params->p_size, ptr, sizeof(params->p_size));
	ptr = dh_unpack_data(&params->g_size, ptr, sizeof(params->g_size));
	params->q_size = 0;
	/*
	 * G is not optional so if we assigned q_size to g_size here when
	 * decoding a new key with Q parameter the length will not match on
	 * well-formed key.
	 */
	if (secret.len != crypto_dh_key_len_no_q(params)) {
		if (len < DH_KPP_SECRET_MIN_SIZE)
			return -EINVAL;
		params->q_size = params->g_size;
		ptr = dh_unpack_data(&params->g_size, ptr, sizeof(params->g_size));
		if (secret.len != crypto_dh_key_len_q(params))
			return -EINVAL;
	}

	/*
	 * Don't permit the buffer for 'key' or 'g' to be larger than 'p', since
	 * some drivers assume otherwise.
	 */
	if (params->key_size > params->p_size ||
	    params->g_size > params->p_size || params->q_size > params->p_size)
		return -EINVAL;

	/* Don't allocate memory. Set pointers to data within
	 * the given buffer
	 */
	params->key = (void *)ptr;
	params->p = (void *)(ptr + params->key_size);
	params->q = (void *)(ptr + params->key_size + params->p_size);
	params->g = (void *)(ptr + params->key_size + params->p_size +
			     params->q_size);

	/*
	 * Don't permit 'p' to be 0.  It's not a prime number, and it's subject
	 * to corner cases such as 'mod 0' being undefined or
	 * crypto_kpp_maxsize() returning 0.
	 */
	if (memchr_inv(params->p, 0, params->p_size) == NULL)
		return -EINVAL;

	/* It is permissible to not provide Q. */
	if (params->q_size == 0)
		params->q = NULL;

	return 0;
}
EXPORT_SYMBOL_GPL(crypto_dh_decode_key);

#undef crypto_dh_encode_key
#undef crypto_dh_decode_key

int crypto_dh_encode_key(char *buf, unsigned int len, const struct dh *old_params)
{
	struct dh_q params = {
		.q_size = 0, .q = NULL,
		.key_size = old_params->key_size,
		.p_size = old_params->p_size,
		.g_size = old_params->g_size,
		.key = old_params->key,
		.p = old_params->p,
		.g = old_params->g,
	};
	return crypto_dh_encode_key_q(buf, len, &params);

}
EXPORT_SYMBOL_GPL(crypto_dh_encode_key);

int crypto_dh_decode_key(const char *buf, unsigned int len, struct dh *old_params)
{
	struct dh_q params;
	int ret = crypto_dh_decode_key_q(buf, len, &params);
	if (ret)
		return ret;
	if (params.q_size)
		return -EOPNOTSUPP;
	old_params->key_size = params.key_size;
	old_params->p_size = params.p_size;
	old_params->g_size = params.g_size;
	old_params->key = params.key;
	old_params->p = params.p;
	old_params->g = params.g;
	return ret;
}
EXPORT_SYMBOL_GPL(crypto_dh_decode_key);
